//***************************************************************************
// Filename:	condWait.c
// Author: Chadd Williams
// Date: 4/1/2010
// Class: CS460
//       Adapted from:
//       http://publib.boulder.ibm.com/iseries/v5r2/ic2924/
//       	index.htm?info/apis/users_78.htm
//
//       This code will start NTHREADS.  Each thread will iterate some number
//       of times printing the values 1-10 to the screen.
//
//   		Iterations are done in lock step, and each thread runs a different
//    		number of iterations.
//
//   		If you see valgrind errors:
//   			http://sourceware.org/bugzilla/show_bug.cgi?id=5171
//***************************************************************************

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>

// these variables are used to track how far each thread has progressed
// gFinished tracks how many threads have run to completion and terminated.
// gIterationCompleted tracks how many threads have completed the current
// 	iteration
// gWaitForAllReady tracks how many threads have reached the ready state.
int gIterationCompleted = 0, gFinished = 0, gWaitForAllReady;

// google PTHREAD_COND_INITIALIZER
pthread_cond_t  gCondIteration  = PTHREAD_COND_INITIALIZER;
pthread_cond_t  gCondReady  = PTHREAD_COND_INITIALIZER;

pthread_mutex_t gIterationsMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t gFinishedMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t gReadyMutex = PTHREAD_MUTEX_INITIALIZER;

#define NTHREADS    10

// RANDOM_SLEEP allows us to randomly sleep for
// some number of microseconds.  This causes the
// thread scheduling (who runs when) to be more
// random and more likely to expose thread race
// conditions.
//#define RANDOM_SLEEP usleep(rand() % 100000 );
#define RANDOM_SLEEP usleep( rand_r(&seedData) % 10000);

//***************************************************************************
// Function: threadFunc
//
// Description: main function for each thread. This function prints the values
//							0 to 9 on each iteration.
//
// Parameters: param - pointer to an int which specifies how many iterations
//						 to run.
//
// Returned: None
//***************************************************************************
void *threadfunc(void *param)
{
	int x;
	int iterations;
	int max = *(int*) param;
	unsigned int seedData;

	/*
	 * Skeleton of this function without any synchronization (just the real
	 * work part)
	 *
	 *
	 * printf("MAX %d",max);
	 * sleep(1);
	 *
	 * for(iterations = 0; iterations < max; iterations++ )
	 * {
	 * 	 // DO WORK
	 * 	 printf("%d ====================\n", pthread_self());
	 * 	 for(x=0; x<10;x++)
	 *   {
	 *   	 printf("%d: %d\n", pthread_self(), x);
	 *   }
	 * }
	 * free((int*)param);
	 * pthread_exit(0);
	 *
	 */

	sleep(1);
	seedData = time(NULL);

	printf("MAX %d SEED %d\n",max, seedData);

	for(iterations = 0; iterations < max; iterations++ )
	{
		RANDOM_SLEEP

		// update the readyMutex (shared memory so hold a lock)
		// this thread has reached the ready state.
		pthread_mutex_lock(&gReadyMutex);
		gWaitForAllReady ++;

		// wait for signal to continue
		// and release the lock so other threads can
		// continue.  The "main" thread can check
		// gWaitForAllReady since each thread releases
		// the lock at pthread_cond_wait
		//
		// The "main" thread can then signal each thread
		// to continue.
  	pthread_cond_wait(&gCondReady, &gReadyMutex);
  	RANDOM_SLEEP
		pthread_mutex_unlock(&gReadyMutex);

		RANDOM_SLEEP

		// DO WORK
		printf("%d ================== Iteration %d\n", pthread_self(), iterations);
		for(x=0; x<10;x++)
		{
			printf("%d: %d\n", pthread_self(), x);
			RANDOM_SLEEP
		}

		RANDOM_SLEEP

		// this thread has finished an iteration
		// use the same idiom as above to update
		// a shared counter and wait for a signal to
		// continue the thread.
		pthread_mutex_lock(&gIterationsMutex);
		gIterationCompleted++;

		// wait for signal to continue
		pthread_cond_wait(&gCondIteration, &gIterationsMutex);
		RANDOM_SLEEP
		pthread_mutex_unlock(&gIterationsMutex);

		RANDOM_SLEEP

	}

	// this thread is finished
	printf("FINISHED %d\n", pthread_self());
	pthread_mutex_lock(&gFinishedMutex);
	gFinished ++;
	pthread_mutex_unlock(&gFinishedMutex);

	RANDOM_SLEEP

	// cleanup dynamic memory
	free((int*)param);
  pthread_exit(0);
}
//***************************************************************************
// Function: waitUntilThreadsFinish
//
// Description: function that manages the thread execution. This function
// 							makes sure that each thread runs one iteration and then
// 							waits for a signal to continue.
//
// Parameters: None
//
// Returned: None
//***************************************************************************
void waitUntilThreadsFinish()
{
	int notReady, notDone;
	unsigned int seedData	= time(NULL);

	printf(" waitUntilThreadsFinish SEED %d\n", seedData);

  // continue running while some thread is active.
	pthread_mutex_lock(&gFinishedMutex);
  while(gFinished != NTHREADS)
  {
  	pthread_mutex_unlock(&gFinishedMutex);
  	notReady = 1;

  	// wait for each thread to get launched and ready
  	while( notReady )
  	{
  		// find out how many threads have achieved
  		// the ready state.  Be sure to access the shared
  		// data only when the mutex is held.
  		//
  		// since all threads run different numbers of
  		// iterations, some threads may have already
  		// terminated. Make sure to only wait for the
  		// active threads to reach the ready state,
  		// hence the "- gFinished"
  		pthread_mutex_lock(&gReadyMutex);
  		if ( gWaitForAllReady == NTHREADS - gFinished )
  		{
  			notReady = 0;
  		}
  		// don't forget to unlock the mutex
  		pthread_mutex_unlock(&gReadyMutex);
  		RANDOM_SLEEP
  	}

  	RANDOM_SLEEP

  	// reset the counter since all threads have reached
  	// the ready state
  	// make sure to hold the mutex while updating
		pthread_mutex_lock(&gReadyMutex);
  	gWaitForAllReady = 0;
		pthread_mutex_unlock(&gReadyMutex);

  	RANDOM_SLEEP
  	// signal all threads to continue
  	pthread_cond_broadcast(&gCondReady);

		RANDOM_SLEEP

  	// count how many threads have completed an iteration
    notDone = 1;
  	while( notDone )
  	{
  		// find out how many threads have achieved
  		// completed an iteration. Be sure to access the shared
  		// data only when the mutex is held.
  		pthread_mutex_lock(&gIterationsMutex);
  		if ( gIterationCompleted == NTHREADS - gFinished )
  		{
  			notDone = 0;
  		}
  		pthread_mutex_unlock(&gIterationsMutex);
  		RANDOM_SLEEP


  	}

  	// once all threads have completed the iteration, reset the counter
		pthread_mutex_lock(&gIterationsMutex);
  	gIterationCompleted = 0;
		pthread_mutex_unlock(&gIterationsMutex);

		RANDOM_SLEEP

		// signal all threads to continue
  	pthread_cond_broadcast(&gCondIteration);
		RANDOM_SLEEP
		pthread_mutex_lock(&gFinishedMutex);
  }
	pthread_mutex_unlock(&gFinishedMutex);

}

//***************************************************************************
// Function: cleanup
//
// Description: deallocate the pthread data
//
// Parameters: None
//
// Returned: None
//***************************************************************************
void cleanup()
{
	// clean up
  pthread_cond_destroy(&gCondIteration);
  pthread_cond_destroy(&gCondReady);
  pthread_mutex_destroy(&gIterationsMutex);
  pthread_mutex_destroy(&gFinishedMutex);
  pthread_mutex_destroy(&gReadyMutex);
}

//***************************************************************************
// Function: main
//
// Description: main function. Launch threads, wait until each thread completes,
//						  join on each thread and then cleanup.
//
// Parameters: None
//
// Returned: None
//***************************************************************************
int main(int argc, char **argv)
{
  int i;
  pthread_t threadid[NTHREADS];
  int *intPtr;
  unsigned int seedData = time(NULL);

  printf("Creating %d threads SEED: %d\n", NTHREADS, seedData);
 	gWaitForAllReady = 0;

 	// launch each thread
  for(i=0; i<NTHREADS; ++i) {
  	intPtr = (int*) malloc(sizeof(int));
  	*intPtr = NTHREADS-i;
    pthread_create(&threadid[i], NULL, threadfunc, intPtr);
  	printf("Thread %d Created with Arg %d\n", threadid[i], *intPtr);

  	//sleep up to 2 seconds
  	usleep( rand_r(&seedData) % 2000000);
  }

  // let the threads run
  waitUntilThreadsFinish();

  // all threads have terminated, join on each one
  for (i=0; i<NTHREADS; ++i) {
  	printf("Wait for thread [%d] and cleanup\n",i);
    pthread_join(threadid[i], NULL);
  }

  cleanup();

  printf("Main completed\n");
  return 0;
}

